/*
 * The Unified Mapping Platform (JUMP) is an extensible, interactive GUI 
 * for visualizing and manipulating spatial features with geometry and attributes.
 *
 * Copyright (C) 2003 Vivid Solutions
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * For more information, contact:
 *
 * Vivid Solutions
 * Suite #1A
 * 2328 Government Street
 * Victoria BC  V8T 5G5
 * Canada
 *
 * (250)385-6040
 * www.vividsolutions.com
 */
package com.vividsolutions.jump.feature;

import java.io.Serializable;
import java.math.BigDecimal;
import java.util.Collection;
import java.util.Date;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.Iterator;

import com.vividsolutions.jts.geom.Geometry;
//import com.vividsolutions.jts.geom.Point;
//import com.vividsolutions.jts.geom.PrecisionModel;


/**
 * Whether an attribute on a feature is a string, geometry, integer, etc.
 *
 * Class modified by Michal MICHAUD on 3 nov. 2004
 * New AttributeTypes are :
 * <ul>
 * <li>LONG (java.lang.Long objects)</li>
 * <li>BOOLEAN (java.lang.Boolean objects)</li>
 * <li>Char (java.lang.String objects with a max length)</li>
 * <li>Decimal (java.Math.BigDecimal objects with limited length and number of decimal digits)</li>
 * <li>Enumeration (java.lang.Object objects selected from a list of possible values</li>
 * </ul>
 * @see com.vividsolutions.jump.feature.FeatureSchema
 */
public class AttributeType implements Serializable {
    private static final long serialVersionUID = -8443945292593638566L;
    // Replace HashMap by TreeMap to get an ordered list of AttributeType(s)
    // when there are many attributes
    private static Map nameToAttributeTypeMap = new TreeMap();
    /**
     * Returns all instances of AttributeType.
     */
    public static Collection allTypes() {
        return nameToAttributeTypeMap.values();
    }

    /** For strings */
    public final static AttributeType STRING = new AttributeType("STRING", String.class);

    /** For spatial data */
    public final static AttributeType GEOMETRY = new AttributeType("GEOMETRY", Geometry.class);

    /**
     * For integer values (32-bit)
     */
    public final static AttributeType INTEGER = new AttributeType("INTEGER", Integer.class);

    /**
     * For long values (64-bit)
     * Integer.class changed to Long.class by Michal MICHAUD on 18 sept. 2004
     * --> allow conversion from Date to INTEGER
     */
    public final static AttributeType LONG = new AttributeType("LONG", Long.class);
    
    /** 
     * For dates 
     * @see java.util.Date 
     */
    public final static AttributeType DATE = new AttributeType("DATE", Date.class);


    /** For double-precision values (64-bit) */
    public final static AttributeType DOUBLE = new AttributeType("DOUBLE", Double.class);
    
    /**
     * For simple-precision values (32-bit)
     * Added by Michael Michaud on 13 nov 2005 for backward compatibility with
     * Agil plugins
     **/
    public final static AttributeType FLOAT = new AttributeType("FLOAT", Float.class);

    /** Experimental */            
    public final static AttributeType OBJECT = new AttributeType("OBJECT", Object.class );
    
    /** For boolean : Added by Michael Michaud on 18 sept 2004*/
    public final static AttributeType BOOLEAN = new AttributeType("BOOLEAN", Boolean.class);

    /**
     * @see http://www.javaworld.com/javaworld/javatips/jw-javatip122.html
     */
    private Object readResolve() {
        return nameToAttributeTypeMap.get(name);
    } 
    
   /**
    * DecimalType example : added by Michael Michaud on 18 sept 2004.
    * Decimal AttributeType are generally created by the user or the JUMPReader.
    */
    public final static AttributeType DECIMAL0601 = new AttributeType.Decimal(6,1);

    // String variable used for true and false inside JUMP (IHM)
    // Internal representation is Boolean, but representation is user
    // user dependant (ex. "oui" instead of true for french user.
    private static String true_ = "true";
    private static String false_ = "false";
    public final static void setTrueString(String ptrue) {true_ = ptrue;}
    public final static String getTrueString() {return true_;}
    public final static void setFalseString(String pfalse) {false_ = pfalse;}
    public final static String getFalseString() {return false_;}
    



    private String name;
    
    // length of the field. For variable length fields, length is set to -1;
    private int length_ = -1;

    protected AttributeType(String name, Class javaClass) {
        this.name = name;
        this.javaClass = javaClass;
        nameToAttributeTypeMap.put(name, this);
    }

    public String toString() {
        return name;
    }
    
   /**
    * Length used to represent the attribute in a fixed length field.
    * Return -1 for variable length attributes (STRING, GEOMETRY,...)
    */
    public int getLength() {return length_;}

    /**
     * Converts a type name to an AttributeType.
     * @param typename the name of the AttributeType to retrieve
     * @return the corresponding AttributeType
     * @throws InvalidAttributeTypeException if the type name is unrecognized
     */
    public final static AttributeType toAttributeType(String name) {
        AttributeType type = (AttributeType) nameToAttributeTypeMap.get(name);
        if (type == null) {
            throw new IllegalArgumentException();
        }
        return type;
    }
    
    /**
     * Returns the Java class of attributes that are said to be of this AttributeType.
     */
	public Class toJavaClass() {
		return javaClass;
	}
	
	private Class javaClass;

    /**
     * Returns the AttributeType of attributes that are instances of the given class,
     * or null if it is not associated with any AttributeType.
     */
    public static AttributeType toAttributeType(Class javaClass) {
        for (Iterator i = allTypes().iterator(); i.hasNext(); ) {
            AttributeType type = (AttributeType) i.next();
            if (type.toJavaClass() == javaClass) {
                return type;
            }
        }
        return null;        
    }
    
   /**
    * FixedLengthString added by Michal MICHAUD on 18 sept. 2004
    */
    public static class Char extends AttributeType {
        public Char (int length) {
            super("CHAR ("+Integer.toString(length)+")", String.class);
            super.length_ = length;
        }
        public int getLength() {return super.length_;}
    }
    
   /**
    * Create a new FixedLengthString AttributeType.
    * added by Michal MICHAUD on 18 sept. 2004
    */
    public static Char createCharAttributeType(int length) {
        return new AttributeType.Char(length);
    }
    
   /**
    * Decimal added by Michal MICHAUD on 18 sept. 2004
    */
    public static class Decimal extends AttributeType {
        private int fractionDigits_ = 2;
        public Decimal (int totalDigits, int fractionDigits) {
            super("DECIMAL (" + Integer.toString(totalDigits) + "," +
                Integer.toString(fractionDigits) +")" , BigDecimal.class);
            super.length_ = totalDigits;
            this.fractionDigits_ = fractionDigits;
        }
    
        public int getFractionDigits() {return fractionDigits_;}
    }
    
   /**
    * Create a new Decimal AttributeType.
    * added by Michal MICHAUD on 18 sept. 2004
    */
    public static Decimal createDecimalAttributeType(int length, int fractionDigits) {
        return new AttributeType.Decimal(length, fractionDigits);
    }
    
    
   /**
    * Enumeration added by Michael Michaud on 18 sept. 2004
    */
    public static class Enumeration extends AttributeType {
        //public Set set;
        /** Constructor*/
        public Enumeration (String name, Set enumeration) {
            super(name, Object.class);
            //this.set = enumeration;
        }
        public Set getEnumeration() {return null;}
        public Object[] getEnumerationArray() {
            if (getEnumeration()==null) return null;
            return (Object[])getEnumeration().toArray(new Object[]{});
        }
        public int intValue(String string) {
            Object[] oo = getEnumerationArray();
            for(int i = 0 ; i < oo.length ; i++) {
                if (string.trim().equalsIgnoreCase(oo[i].toString().trim())) {
                    return i;
                }
            }
            return -1;
        }
	public String toString() {
		StringBuffer sb = new StringBuffer(super.toString() + " [");
		Object[] oo = getEnumerationArray();
		for (int i = 0 ; i < oo.length ; i++) {
            sb.append("\""+oo[i].toString()+"\",");
        }
		sb.setCharAt(sb.length()-1,']');
		return sb.toString();
	}
    }
    
   /**
    * Create a new Enumeration with a name and an Object Array.
    * added by Michal MICHAUD on 18 sept. 2004
    */
    public static Enumeration createEnum(String name, final Object[] array) {
        final Set pset = new LinkedHashSet();
        for (int i = 0 ; i < array.length ; i++) {
            pset.add(array[i]);
        }
        /** Enumeration creator overload the getEnumeration method.*/
        return new AttributeType.Enumeration(name, pset) {
            public Set getEnumeration() {return pset;}
        };
    }
    
}
