/*
 * The Unified Mapping Platform (JUMP) is an extensible, interactive GUI
 * for visualizing and manipulating spatial features with geometry and attributes.
 *
 * Copyright (C) 2003 Vivid Solutions
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * For more information, contact:
 *
 * Vivid Solutions
 * Suite #1A
 * 2328 Government Street
 * Victoria BC  V8T 5G5
 * Canada
 *
 * (250)385-6040
 * www.vividsolutions.com
 */
package com.vividsolutions.jump.workbench.ui.plugin;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.io.ParseException;
import com.vividsolutions.jts.io.WKTReader;
import com.vividsolutions.jts.util.Assert;
// Added on 10 sept. 2004 by Michal MICHAUD
import java.util.Set;

import com.vividsolutions.jump.feature.AttributeType;

import com.vividsolutions.jump.feature.BasicFeature;
import com.vividsolutions.jump.feature.Feature;
import com.vividsolutions.jump.feature.FeatureDataset;
import com.vividsolutions.jump.feature.FeatureSchema;
import com.vividsolutions.jump.util.FlexibleDateParser;
import com.vividsolutions.jump.util.StringUtil;
import com.vividsolutions.jump.workbench.WorkbenchContext;
import com.vividsolutions.jump.workbench.model.CategoryEvent;
import com.vividsolutions.jump.workbench.model.FeatureEvent;
import com.vividsolutions.jump.workbench.model.Layer;
import com.vividsolutions.jump.workbench.model.LayerEvent;
import com.vividsolutions.jump.workbench.model.LayerEventType;
import com.vividsolutions.jump.workbench.model.LayerListener;
import com.vividsolutions.jump.workbench.model.LayerManager;
import com.vividsolutions.jump.workbench.model.LayerManagerProxy;
import com.vividsolutions.jump.workbench.model.Layerable;
import com.vividsolutions.jump.workbench.plugin.AbstractPlugIn;
import com.vividsolutions.jump.workbench.plugin.EnableCheckFactory;
import com.vividsolutions.jump.workbench.plugin.MultiEnableCheck;
import com.vividsolutions.jump.workbench.plugin.PlugInContext;
import com.vividsolutions.jump.workbench.ui.LayerNamePanel;
import com.vividsolutions.jump.workbench.ui.LayerNamePanelListener;
import com.vividsolutions.jump.workbench.ui.LayerNamePanelProxy;
import com.vividsolutions.jump.workbench.ui.SchemaPanel;
import com.vividsolutions.jump.workbench.ui.TreeLayerNamePanel;
import com.vividsolutions.jump.workbench.ui.WorkbenchFrame;
import com.vividsolutions.jump.workbench.ui.cursortool.editing.EditingPlugIn;
import com.vividsolutions.jump.workbench.ui.images.IconLoader;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

// added on 10 sept 2004 by Michael Michaud
import java.math.BigDecimal;

import java.text.DateFormat;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import javax.swing.ImageIcon;
import javax.swing.JInternalFrame;
import javax.swing.JOptionPane;
import javax.swing.event.InternalFrameAdapter;
import javax.swing.event.InternalFrameEvent;

/**
 *  The PlugIn has been modified by Michal MICHAUD on 3 nov. 2004 to handle
 *  new AttributeTypes
 *  @see com.vividsolutions.jump.feature.AttributeType
 */
public class ViewSchemaPlugIn extends AbstractPlugIn {
    private static final String KEY = ViewSchemaPlugIn.class + " - FRAME";
    private EditingPlugIn editingPlugIn;
    private GeometryFactory factory = new GeometryFactory();
    private WKTReader wktReader = new WKTReader(factory);
    private FlexibleDateParser dateParser = new FlexibleDateParser();
    private DateFormat dateFormatter = DateFormat.getDateInstance();

    public ViewSchemaPlugIn(EditingPlugIn editingPlugIn) {
        this.editingPlugIn = editingPlugIn;
    }

    public String getName() {
        return "View / Edit Schema";
    }

    private void applyChanges(final Layer layer, final SchemaPanel panel)
        throws Exception {
        if (!panel.isModified()) {
            //User just pressed the Apply button even though he made no edits.
            //Don't truncate the undo history; instead, exit. [Jon Aquino]
            return;
        }

        if (panel.validateInput() != null) {
            throw new Exception(panel.validateInput());
        }

        panel.getModel().removeBlankRows();

        FeatureSchema newSchema = new FeatureSchema();

        for (int i = 0; i < panel.getModel().getRowCount(); i++) {
            newSchema.addAttribute(panel.getModel().get(i).getName(),
                panel.getModel().get(i).getType());
        }

        List originalFeatures = layer.getFeatureCollectionWrapper().getFeatures();
        ArrayList tempFeatures = new ArrayList();

        //Two-phase commit. Phase 1: check that no conversion errors occur. [Jon Aquino]
        for (Iterator i = layer.getFeatureCollectionWrapper().iterator();
                i.hasNext();) {
            Feature feature = (Feature) i.next();
            tempFeatures.add(convert(feature, panel, newSchema));
        }

        //Phase 2: commit. [Jon Aquino]
        for (int i = 0; i < originalFeatures.size(); i++) {
            Feature originalFeature = (Feature) originalFeatures.get(i);
            Feature tempFeature = (Feature) tempFeatures.get(i);

            //Modify existing features rather than creating new features, because
            //there may be references to the existing features (e.g. Attribute Viewers).
            //[Jon Aquino]            
            originalFeature.setSchema(tempFeature.getSchema());
            originalFeature.setAttributes(tempFeature.getAttributes());
        }

        //Non-undoable. [Jon Aquino]
        layer.getLayerManager().getUndoableEditReceiver().getUndoManager()
             .discardAllEdits();
        layer.setFeatureCollection(new FeatureDataset(originalFeatures,
                newSchema));
        layer.fireLayerChanged(LayerEventType.METADATA_CHANGED);
        panel.markAsUnmodified();
    }

    private Feature convert(Feature oldFeature, SchemaPanel panel,
        FeatureSchema newSchema) throws ConversionException {
        Feature newFeature = new BasicFeature(newSchema);
        for (int i = 0; i < panel.getModel().getRowCount(); i++) {
            if (panel.getModel().get(i).getOriginalIndex() == -1) {
                newFeature.setAttribute(i,
                    (panel.getModel().get(i).getType() == AttributeType.GEOMETRY)
                    ? oldFeature.getGeometry() : null);
            } else {
                newFeature.setAttribute(i,
                    convert(oldFeature.getAttribute(panel.getModel().get(i)
                                                         .getOriginalIndex()),
                        oldFeature.getSchema().getAttributeType(panel.getModel()
                                                                     .get(i)
                                                                     .getOriginalIndex()),
                        newFeature.getSchema().getAttributeType(i),
                        panel.getModel().get(i).getName(),
                        panel.isForcingInvalidConversionsToNull()));
            }
        }
        return newFeature;
    }

    private String limitLength(String s) {
        //Limit length of values reported in error messages -- WKT is potentially large.
        //[Jon Aquino]
        return StringUtil.limitLength(s, 30);
    }

    private Object convert(Object oldValue, AttributeType oldType,
        AttributeType newType, String name,
        boolean forcingInvalidConversionsToNull) throws ConversionException {
        try {
            if (oldValue == null) {
                return (newType == AttributeType.GEOMETRY)
                ? factory.createPoint((Coordinate)null) : null;
            }
            
            //////////////////////////////////////
            // CONVERT A STRING
            //////////////////////////////////////
            if (oldType.toJavaClass().equals(java.lang.String.class)) {
                return convertString(oldValue, newType, name, forcingInvalidConversionsToNull);
            }

            //////////////////////////////////////
            // CONVERT AN INTEGER
            //////////////////////////////////////
            else if (oldType == AttributeType.INTEGER) {
                return convertInteger(oldValue, newType, name, forcingInvalidConversionsToNull);
            }
            
            //////////////////////////////////////
            // CONVERT A LONG
            //////////////////////////////////////
            else if (oldType == AttributeType.LONG) {
                return convertLong(oldValue, newType, name, forcingInvalidConversionsToNull);
            }

            //////////////////////////////////////
            // CONVERT A DOUBLE
            //////////////////////////////////////
            if (oldType == AttributeType.DOUBLE) {
                return convertDouble(oldValue, newType, name, forcingInvalidConversionsToNull);
            }
            if (oldType instanceof AttributeType.Decimal) {
                return convertDecimal(oldValue, newType, name, forcingInvalidConversionsToNull);
            }

            //////////////////////////////////////
            // CONVERT A GEOMETRY
            //////////////////////////////////////
            if (oldType == AttributeType.GEOMETRY) {
                return convertGeometry(oldValue, newType, name, forcingInvalidConversionsToNull);
            }

            if (oldType == AttributeType.DATE) {
                return convertDate(oldValue, newType, name, forcingInvalidConversionsToNull);
            }
            //////////////////////////////////////
            // CONVERT A BOOLEAN
            //////////////////////////////////////
            if (oldType == AttributeType.BOOLEAN) {
                return convertBoolean(oldValue, newType, name, forcingInvalidConversionsToNull);
            }

            if (oldType instanceof AttributeType.Enumeration) {
                return convertEnumeration(oldValue, oldType, newType, name, forcingInvalidConversionsToNull);
            }

            Assert.shouldNeverReachHere(oldType.toString() + "-->" + newType.toString());

            return null;
        } catch (ConversionException e) {
            if (forcingInvalidConversionsToNull) {
                return (newType == AttributeType.GEOMETRY)
                ? factory.createPoint((Coordinate)null) : null;
            }
            throw e;
        }
    }
    
    private Object convertNull(AttributeType newType, String name,
        boolean forcingInvalidConversionsToNull) throws ConversionException {
                return (newType == AttributeType.GEOMETRY)
                ? factory.createPoint((Coordinate)null) : null;
    }
    
    private Object convertString(Object oldValue, AttributeType newType,
                String name, boolean forcingInvalidConversionsToNull)
                                                    throws ConversionException {
        String oldString = (String) oldValue;

        if (newType.toJavaClass().equals(java.lang.String.class)) {
            int length = newType.getLength();
            if (length < 0 || oldString.length()<=length) return oldString;
            else return oldString.substring(0, length);
        }
        else if (newType == AttributeType.INTEGER) {
            try {return new Integer(oldString);}
            catch (NumberFormatException e) {
                throw new ConversionException("Cannot convert to integer: \"" +
                limitLength(oldString.toString()) + "\" (" + name + ")");
            }
        }
        else if (newType == AttributeType.LONG) {
            try {return new Long(oldString);}
            catch (NumberFormatException e) {
                throw new ConversionException("Cannot convert to integer: \"" +
                limitLength(oldString.toString()) + "\" (" + name + ")");
            }
        }
        else if (newType == AttributeType.DOUBLE) {
            try {return new Double(oldString);}
            catch (NumberFormatException e) {
                throw new ConversionException("Cannot convert to double: \"" +
                limitLength(oldString.toString()) + "\" (" + name + ")");
            }
        }
        else if (newType == AttributeType.GEOMETRY) {
            try {return wktReader.read(oldString);}
            catch (ParseException e) {throw new ConversionException(
                  "Cannot convert to geometry: \"" +
                  limitLength(oldString.toString()) + "\" (" + name + ")");
            }
        }
        else if (newType == AttributeType.DATE) {
            try {return dateParser.parse(oldString, false);}
            catch (java.text.ParseException e) {
                throw new ConversionException("Cannot convert to date: \"" +
                limitLength(oldString.toString()) + "\" (" + name + ")");
            }
        }
        else if (newType == AttributeType.BOOLEAN) {
            if (oldString.length()==0) return new Boolean(false);
            char c = oldString.toLowerCase().charAt(0);
            if (c=='t' || c=='y' || c=='1' || c=='o') return new Boolean(true);
            else if (c=='f' || c=='n' || c=='0') return new Boolean(false);
            else throw new ConversionException("Cannot convert to boolean: \"" +
                limitLength(oldString) + "\" (" + name + ")");
        }
        else if (newType instanceof AttributeType.Decimal) {
            try {
                int fractionDigits = ((AttributeType.Decimal)newType).getFractionDigits();
                double mult = 1.0;
                for (int i=0 ; i<fractionDigits ; i++) mult*=10;
                return new BigDecimal(Math.rint(new Double(oldString).doubleValue()*mult)/mult);
            }
            catch (NumberFormatException e) {
                throw new ConversionException("Cannot convert to Decimal: \"" +
                    limitLength(oldString) + "\" (" + name + ")");
            }
        }
        else if (newType instanceof AttributeType.Enumeration) {
            Set enumeration = ((AttributeType.Enumeration)newType).getEnumeration();
            for (Iterator it = enumeration.iterator() ; it.hasNext() ; ) {
                Object o = it.next();
                if(oldString.trim().equals(o.toString())) {
                    return o;
                }
            }
            enumeration.add(oldString.trim());
            return oldString;
        }
        else return null;
    }
    
    private Object convertInteger(Object oldValue, AttributeType newType,
                String name, boolean forcingInvalidConversionsToNull)
                                                    throws ConversionException {
        
        int oldInteger = ((Integer) oldValue).intValue();

        if (newType.toJavaClass().equals(java.lang.String.class)) {
            if(newType == AttributeType.STRING) return "" + oldInteger;
            else return ("" + oldInteger).substring(0, newType.getLength());
        }
        else if (newType == AttributeType.INTEGER) {
            return oldValue;
        }
        else if (newType == AttributeType.LONG) {
            return new Long((long)oldInteger);
        }
        else if (newType == AttributeType.DOUBLE) {
                return new Double((double)oldInteger);
        }
        else if (newType instanceof AttributeType.Decimal) {
            int length = newType.getLength();
            return new BigDecimal(Integer.toString(oldInteger).substring(0,length));
        }
        else if (newType == AttributeType.GEOMETRY) {
            throw new ConversionException(
                "Cannot convert to geometry: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.DATE) {
            throw new ConversionException(
                "Cannot convert to date: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.BOOLEAN) {
            if (oldInteger==0) return new Boolean(false);
            else return new Boolean(true);
        }
        else if (newType instanceof AttributeType.Enumeration) {
            try {
                return ((AttributeType.Enumeration)newType).getEnumerationArray()[(int)oldInteger];
            }
            catch(ArrayIndexOutOfBoundsException e) {
                throw new ConversionException("Cannot convert to enumeration: \"" +
                    limitLength(Integer.toString(oldInteger)) + "\" (" + name + ")");
            }
        }
        else return null;
    }
    
    private Object convertLong(Object oldValue, AttributeType newType,
                String name, boolean forcingInvalidConversionsToNull)
                                                    throws ConversionException {
        
        long oldLong = ((Long) oldValue).longValue();

        if (newType.toJavaClass().equals(java.lang.String.class)) {
            if(newType == AttributeType.STRING) return "" + oldLong;
            else return ("" + oldLong).substring(0, newType.getLength());
        }
        else if (newType == AttributeType.INTEGER) {
            return new Integer((int)oldLong);
        }
        else if (newType == AttributeType.LONG) {
            return oldValue;
        }
        else if (newType == AttributeType.DOUBLE) {
                return new Double((double)oldLong);
        }
        else if (newType instanceof AttributeType.Decimal) {
            int length = newType.getLength();
            return new BigDecimal(Long.toString(oldLong).substring(0,length));
        }
        else if (newType == AttributeType.GEOMETRY) {
            throw new ConversionException(
                "Cannot convert to geometry: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.DATE) {
            return new Date(oldLong);
        }
        else if (newType == AttributeType.BOOLEAN) {
            if (oldLong==0) return new Boolean(false);
            else return new Boolean(true);
        }
        else if (newType instanceof AttributeType.Enumeration) {
            try {
                return ((AttributeType.Enumeration)newType).getEnumerationArray()[(int)oldLong];
            }
            catch(ArrayIndexOutOfBoundsException e) {
                throw new ConversionException("Cannot convert to enumeration: \"" +
                    limitLength(Long.toString(oldLong)) + "\" (" + name + ")");
            }
        }
        else return null;
    }
    
    private Object convertDouble(Object oldValue, AttributeType newType,
                String name, boolean forcingInvalidConversionsToNull)
                                                    throws ConversionException {
        
        double oldDouble = ((Double) oldValue).doubleValue();

        if (newType.toJavaClass().equals(java.lang.String.class)) {
            if(newType == AttributeType.STRING) return "" + oldDouble;
            else return ("" + oldDouble).substring(0, newType.getLength());
        }
        else if (newType == AttributeType.INTEGER) {
            return new Integer((int) oldDouble);
        }
        else if (newType == AttributeType.LONG) {
            return new Long((long) oldDouble);
        }
        else if (newType.toJavaClass().equals(java.lang.Double.class)) {
            if (newType == AttributeType.DOUBLE) return oldValue;
            else {
                int fd = ((AttributeType.Decimal)newType).getFractionDigits();
                double mult = 1.0;
                for (int i = 0 ; i < fd ; i++) mult *= 10;
                return new Double(Math.rint(oldDouble*mult)/mult);
            }
        }
        else if (newType instanceof AttributeType.Decimal) {
            int fd = ((AttributeType.Decimal)newType).getFractionDigits();
            double mult = 1.0;
            for (int i = 0 ; i < fd ; i++) mult *= 10;
            return new BigDecimal(Math.rint(oldDouble*mult)/mult);
        }
        else if (newType == AttributeType.GEOMETRY) {
            throw new ConversionException(
                "Cannot convert to geometry: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.DATE) {
            return new Date((long)oldDouble);
            //throw new ConversionException("Cannot convert to date: \"" +
            //    limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.BOOLEAN) {
            if (oldDouble==0.0) return new Boolean(false);
            else return new Boolean(true);
        }
        else if (newType instanceof AttributeType.Enumeration) {
            throw new ConversionException(
                "Cannot convert to enumeration: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else return null;
    }
    
    private Object convertDecimal(Object oldValue, AttributeType newType,
                String name, boolean forcingInvalidConversionsToNull)
                                                    throws ConversionException {
        
        java.math.BigDecimal oldBD = (BigDecimal) oldValue;

        if (newType.toJavaClass().equals(java.lang.String.class)) {
            if(newType == AttributeType.STRING) return oldBD.toString();
            else return oldBD.toString().substring(0, newType.getLength());
        }
        else if (newType == AttributeType.INTEGER) {
            return new Integer(oldBD.intValue());
        }
        else if (newType == AttributeType.LONG) {
            return new Long(oldBD.longValue());
        }
        else if (newType == AttributeType.DOUBLE) {
            return new Double(oldBD.doubleValue());
        }
        else if (newType instanceof AttributeType.Decimal) {
            int fd = ((AttributeType.Decimal)newType).getFractionDigits();
            return oldBD.setScale(fd, java.math.BigDecimal.ROUND_HALF_UP);
        }
        else if (newType == AttributeType.GEOMETRY) {
            throw new ConversionException(
                "Cannot convert to geometry: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.DATE) {
            return new Date(oldBD.longValue());
            //throw new ConversionException("Cannot convert to date: \"" +
            //    limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.BOOLEAN) {
            if (oldBD.intValue()==0) return new Boolean(false);
            else return new Boolean(true);
        }
        else if (newType instanceof AttributeType.Enumeration) {
            throw new ConversionException(
                "Cannot convert to enumeration: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else return null;
    }
    
    
    private Object convertGeometry(Object oldValue, AttributeType newType,
                String name, boolean forcingInvalidConversionsToNull)
                                                    throws ConversionException {
        
        Geometry geom = (Geometry) oldValue;

        if (newType.toJavaClass().equals(java.lang.String.class)) {
            if(newType == AttributeType.STRING) return geom.toString();
            else return (geom.toString()).substring(0, newType.getLength());
        }
        else if (newType == AttributeType.INTEGER) {
            throw new ConversionException(
                "Cannot convert to integer: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.LONG) {
            throw new ConversionException(
                "Cannot convert to long: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.DOUBLE) {
            throw new ConversionException("Cannot convert to double: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType instanceof AttributeType.Decimal) {
            throw new ConversionException("Cannot convert to decimal: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.GEOMETRY) {
            return geom;
        }
        else if (newType == AttributeType.DATE) {
            throw new ConversionException("Cannot convert to date: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType.toString().equals("BOOLEAN")) {
            if (geom.isEmpty() || !geom.isSimple()) return new Boolean(false);
            else return new Boolean(true);
        }
        else if (newType instanceof AttributeType.Enumeration) {
            throw new ConversionException(
                "Cannot convert to enumeration: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else return null;
    }
    
    private Object convertDate(Object oldValue, AttributeType newType,
                String name, boolean forcingInvalidConversionsToNull)
                                                    throws ConversionException {
        
        Date oldDate = (Date) oldValue;

        if (newType.toJavaClass().equals(java.lang.String.class)) {
            if(newType == AttributeType.STRING) return oldDate.toString();
            else if(newType.getLength()<4) {return null;}
            else if (newType.getLength()==4) {
                return new java.text.SimpleDateFormat("yyyy").format(oldDate);
            }
            else if (newType.getLength()==5) {
                return new java.text.SimpleDateFormat("yyyy").format(oldDate);
            }
            else if (newType.getLength()==6) {
                return new java.text.SimpleDateFormat("yyMMdd").format(oldDate);
            }
            else if (newType.getLength()==7) {
                return new java.text.SimpleDateFormat("ddMMyy").format(oldDate);
            }
            else if (newType.getLength()==8) {
                return new java.text.SimpleDateFormat("yyyyMMdd").format(oldDate);
            }
            else if (newType.getLength()==9) {
                return new java.text.SimpleDateFormat("ddMMyyyy").format(oldDate);
            }
            else if (newType.getLength()==10) {
                return new java.text.SimpleDateFormat("yyyy/MM/dd").format(oldDate);
            }
            else if (newType.getLength()==11) {
                return new java.text.SimpleDateFormat("dd/MM/yyyy").format(oldDate);
            }
            else if (newType.getLength()==12) {
                return new java.text.SimpleDateFormat("MMM dd, yyyy").format(oldDate);
            }
            else if (newType.getLength()==13) {
                return new java.text.SimpleDateFormat("dd MMM yyyy").format(oldDate);
            }
            else return new java.text.SimpleDateFormat("yyyy/MM/dd").format(oldDate);
        }
        // 
        else if (newType == AttributeType.INTEGER) {
            return new Integer(oldDate.getYear());
        }
        else if (newType == AttributeType.LONG) {
            return new Long((long) oldDate.getTime());
        }
        else if (newType == AttributeType.DOUBLE) {
            return new Double((double)oldDate.getTime());
        }
        else if (newType instanceof AttributeType.Decimal) {
            throw new ConversionException("Cannot convert to decimal: \"" +
                limitLength(oldDate.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.GEOMETRY) {
            throw new ConversionException(
                "Cannot convert to geometry: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.DATE) {
            return oldDate;
        }
        else if (newType == AttributeType.BOOLEAN) {
            throw new ConversionException(
                "Cannot convert to boolean: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType instanceof AttributeType.Enumeration) {
            throw new ConversionException(
                "Cannot convert to enumeration: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else return null;
    }
    
    private Object convertBoolean(Object oldValue, AttributeType newType,
                String name, boolean forcingInvalidConversionsToNull)
                                                    throws ConversionException {
        
        boolean b = ((Boolean)oldValue).booleanValue();

        if (newType.toJavaClass().equals(java.lang.String.class)) {
            if(newType == AttributeType.STRING) {
                if (b) return AttributeType.getTrueString();
                else return AttributeType.getFalseString();
            }
            else {
                if (b) return (AttributeType.getTrueString().substring(0, newType.getLength()));
                else return (AttributeType.getFalseString().substring(0, newType.getLength()));
            }
        }

        else if (newType == AttributeType.INTEGER) {
            return b?new Integer(1):new Integer(0);
        }
        else if (newType == AttributeType.LONG) {
            return b?new Long(1):new Long(0);
        }
        else if (newType == AttributeType.DOUBLE) {
            return b?new Double(1.0):new Double(0.0);
        }
        else if (newType instanceof AttributeType.Decimal) {
            return b?new BigDecimal("1"):new BigDecimal("0");
        }
        else if (newType == AttributeType.GEOMETRY) {
            throw new ConversionException("Cannot convert to geometry: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.DATE) {
            throw new ConversionException("Cannot convert to date: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType.toString().equals("BOOLEAN")) {
            return oldValue;
        }
        else if (newType instanceof AttributeType.Enumeration) {
            throw new ConversionException("Cannot convert to enumeration: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else return null;
    }
    
    private Object convertEnumeration(Object oldValue,
                AttributeType oldType, AttributeType newType,
                String name, boolean forcingInvalidConversionsToNull)
                                                    throws ConversionException {
        
        //Object oldObject = ((Boolean)oldValue).booleanValue();

        if (newType.toJavaClass().equals(java.lang.String.class)) {
            if(newType == AttributeType.STRING) {
                return oldValue.toString();
            }
            else {
                return oldValue.toString().substring(0, newType.getLength());
            }
        }
        else if (newType == AttributeType.INTEGER) {
            Object[] enumeration = ((AttributeType.Enumeration)oldType).getEnumerationArray();
            for (int i = 0 ; i < enumeration.length ; i++) {
                if (oldValue.equals(enumeration[i])) return new Integer(i);
            }
            return new Integer[-1];
        }
        else if (newType == AttributeType.LONG) {
            Object[] enumeration = ((AttributeType.Enumeration)oldType).getEnumerationArray();
            for (int i = 0 ; i < enumeration.length ; i++) {
                if (oldValue.equals(enumeration[i])) return new Long(i);
            }
            return new Long[-1];
        }
        else if (newType == AttributeType.DOUBLE) {
            throw new ConversionException("Cannot convert to double: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType instanceof AttributeType.Decimal) {
            throw new ConversionException("Cannot convert to decimal: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.GEOMETRY) {
            throw new ConversionException("Cannot convert to geometry: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.DATE) {
            throw new ConversionException("Cannot convert to date: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType == AttributeType.BOOLEAN) {
            throw new ConversionException("Cannot convert to boolean: \"" +
                limitLength(oldValue.toString()) + "\" (" + name + ")");
        }
        else if (newType instanceof AttributeType.Enumeration) {
            Object[] oldEnumeration = ((AttributeType.Enumeration)oldType).getEnumerationArray();
            Object[] newEnumeration = ((AttributeType.Enumeration)newType).getEnumerationArray();
            for (int i = 0 ; i < oldEnumeration.length ; i++) {
                if (oldValue.equals(oldEnumeration[i])) {
                    try {
                        return newEnumeration[i];
                    }
                    catch(ArrayIndexOutOfBoundsException e) {return null;}
                }
            }
            return null;
        }
        else return null;
    }


    private void commitEditsInProgress(final SchemaPanel panel) {
        //Skip if nothing is being edited, otherwise may get false positive. [Jon Aquino]
        if (panel.getTable().getEditingRow() != -1) {
            //If user is in the middle of editing a field name, call stopCellEditing
            //so that new field name is committed (if valid) or an error is recorded
            //(if invalid). [Jon Aquino]
            panel.getTable()
                 .getCellEditor(panel.getTable().getEditingRow(),
                panel.getTable().getEditingColumn()).stopCellEditing();
        }
    }

    public boolean execute(PlugInContext context) throws Exception {
        reportNothingToUndoYet(context);

        //Can't simply use Blackboard#get(key, default) because default requires that
        //we create a new EditSchemaFrame, and we don't want to do this unless we
        //have to because the EditSchemaFrame constructor modifies the blackboard. 
        //Result: Executing this plug-in twice creates two frames, even if we don't close
        //the first. [Jon Aquino]
        if (frame(context) == null) {
            context.getSelectedLayer(0).getBlackboard().put(KEY,
                new EditSchemaFrame(context.getWorkbenchFrame(),
                    context.getSelectedLayer(0), editingPlugIn));
        }

        frame(context).surface();

        return true;
    }

    private EditSchemaFrame frame(PlugInContext context) {
        return (EditSchemaFrame) context.getSelectedLayer(0).getBlackboard()
                                        .get(KEY);
    }

    public static MultiEnableCheck createEnableCheck(
        final WorkbenchContext workbenchContext) {
        EnableCheckFactory checkFactory = new EnableCheckFactory(workbenchContext);

        return new MultiEnableCheck().add(checkFactory.createWindowWithLayerNamePanelMustBeActiveCheck())
                                     .add(checkFactory.createExactlyNLayersMustBeSelectedCheck(
                1));
    }

    private static class ConversionException extends Exception {
        public ConversionException(String message) {
            super(message);
        }
    }
    
    public static final ImageIcon ICON = IconLoader.icon("Object.gif");

    private class EditSchemaFrame extends JInternalFrame
        implements LayerNamePanelProxy, LayerNamePanel, LayerManagerProxy {
        private LayerManager layerManager;
        private Layer layer;
        private WorkbenchFrame workbenchFrame;

        public EditSchemaFrame(final WorkbenchFrame workbenchFrame,
            final Layer layer, EditingPlugIn editingPlugIn) {
            this.layer = layer;
            this.workbenchFrame = workbenchFrame;
            layer.getBlackboard().put(KEY, this);

            this.layerManager = layer.getLayerManager();
            addInternalFrameListener(new InternalFrameAdapter() {
                    public void internalFrameClosed(InternalFrameEvent e) {
                        layer.getBlackboard().put(KEY, null);
                    }
                });

            final SchemaPanel panel = new SchemaPanel(layer, editingPlugIn,
                    workbenchFrame.getContext());
            setResizable(true);
            setClosable(true);
            setMaximizable(true);
            setIconifiable(true);
            getContentPane().setLayout(new BorderLayout());
            getContentPane().add(panel, BorderLayout.CENTER);
            setSize(500, 300);
            updateTitle(layer);
            layer.getLayerManager().addLayerListener(new LayerListener() {
                    public void categoryChanged(CategoryEvent e) {
                    }

                    public void featuresChanged(FeatureEvent e) {
                    }

                    public void layerChanged(LayerEvent e) {
                        updateTitle(layer);
                    }
                });
            setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
            panel.add(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        try {
                            commitEditsInProgress(panel);
                            applyChanges(layer, panel);
                        } catch (Exception x) {
                            workbenchFrame.handleThrowable(x);
                        }
                    }
                });
            addInternalFrameListener(new InternalFrameAdapter() {
                    public void internalFrameClosing(InternalFrameEvent e) {
                        commitEditsInProgress(panel);

                        if (!layer.isEditable() || !panel.isModified()) {
                            dispose();

                            return;
                        }

                        switch (JOptionPane.showConfirmDialog(EditSchemaFrame.this,
                            "Apply changes to the schema?", "JUMP",
                            JOptionPane.YES_NO_CANCEL_OPTION,
                            JOptionPane.WARNING_MESSAGE)) {
                        case JOptionPane.YES_OPTION:

                            try {
                                applyChanges(layer, panel);
                            } catch (Exception x) {
                                workbenchFrame.handleThrowable(x);

                                return;
                            }

                            dispose();

                            return;

                        case JOptionPane.NO_OPTION:
                            dispose();

                            return;

                        case JOptionPane.CANCEL_OPTION:
                            return;

                        default:
                            Assert.shouldNeverReachHere();
                        }
                    }
                });
        }

        private void updateTitle(Layer layer) {
            setTitle((layer.isEditable() ? "Edit" : "View") + " Schema: " +
                layer.getName());
        }

        public LayerManager getLayerManager() {
            return layerManager;
        }

        public Layer chooseEditableLayer() {
            return TreeLayerNamePanel.chooseEditableLayer(this);
        }

        public void surface() {
            if (!workbenchFrame.hasInternalFrame(this)) {
                workbenchFrame.addInternalFrame(this, false, true);
            }

            workbenchFrame.activateFrame(this);
            moveToFront();
        }

        public LayerNamePanel getLayerNamePanel() {
            return this;
        }

        public Collection getSelectedCategories() {
            return new ArrayList();
        }

        public Layer[] getSelectedLayers() {
            return new Layer[] { layer };
        }

        public Collection selectedNodes(Class c) {
            if (!Layerable.class.isAssignableFrom(c)) {
                return new ArrayList();
            }

            return Arrays.asList(getSelectedLayers());
        }

        public void addListener(LayerNamePanelListener listener) {}
        public void removeListener(LayerNamePanelListener listener) {}
    }
}
